/*****************************************************************************/
/*
    Anny's Lab 1.0
    made by m@rtlin, VII-IX 2012
    martlin@seznam.cz, http://www.martlins-web.tode.cz/
    Licence: GNU/GPL, Open Source

    Multiplatformní (primárně určena pro GNU/Linux) textová bludišťovka volitelně využívající knihovnu ncurses
    Přeji Hodně zábavy!
*/
/*
    Použití knihovny ncurses:
    LINUX:
        Knihovnu ncurses nainstalujete například (záleží na distribuci):
				# apt-get install ncurses-dev

        Pokud kompilujete napřímo (použitím g++) budete muset zadat přepínač -lncurses,
        nebo zakomentovat řádek "#define USE_CURSES" v souboru declarations.cpp a knihovnu ncurses nepoužívat

        Pokud kompilujete použtím make, je ve výchozím nastavení zapnuté její použití.
        Pokud ji nechcete používat, mustíte volat make s parametrem: "make no-use-curses=1"
    WINDOWS:
        Na windows se mi ji bohužel nepodařilo sehnat a je proto nutné rovnou zakomentovat řádek #define USE_CURSES" v souboru declarations.cpp a knihovnu ncurses nepoužívat
*/
/*
    Soubor declarations.cpp:

    Deklarace preprocesorových maker a použivaných strutkur, unionů a namespaců.
*/
/*****************************************************************************/
#ifndef _al_declarations
#define _al_declarations

#include <cstdlib>
#include <cstdio>
#include <ctime>

/* --- Životně důležitá makra -----------------------------------------------*/
//Verze programu
#ifndef VERSION
    #define VERSION "1.0"
#endif

//Používat vylepšené UI využívající knihovnu ncurses? Jestli ne, zakomentovat "#define USE_CURSES" nebo kompilovat s přepínačem - viz Makefile
#if !defined(USE_CURSES) && !defined(NO_USE_CURSES)
    #define USE_CURSES

		#ifdef USE_CURSES
			#error Na Linuxu kompilujte s prepinacem -lncurses a zakomentujte tento radek, jinak vam vypadne tuna chyb,\
			 nebo vypnete knihovnu ncurses zakomentovanim #define USE_CURSES
		#endif
#endif

//Počet směrů, kterými se lze z daného bodu/uzlu vydat
#define DIRECTC 4

/* --- Deklarace polodynamických maker --------------------------------------*/
//Příkaz pro smazání stránky v závislosti na platformě
#if (defined(__linux__) && __linux__==1) || (defined (__unix__) && __unix__==1)
    #define OSLINUX
	#define CLEARCMD "clear"
#elif defined(__MSDOS__) || defined(__WIN__) || defined(__WIN32__) || defined(__WIN64__) || defined(__WINDOWS__)
    #define OSWINDOWS
	#define CLEARCMD "cls"
#else
	#define OSOTHER
    #define CLEARCMD "cls"
#endif

#define IS_MAP_SMALL ((gameParams::height<gameParams::width) ? (gameParams::height<5 || gameParams::width<7) : (gameParams::height<7 || gameParams::width<5))

//Deklarace příkazů UI. Buďto se použíjí ty z ncurses, nebo klasické céčkové.
//Pozor, PRINTF nekončí závorkou, ERRPRINTF bere jeden argumrnt- chybovou hlášku.
#ifdef USE_CURSES
	//pokud nechcete vyuzivat knihovnu ncurses, zakomentujte radek "#define USE_CURSES" vyse
	#include <ncurses.h>
    #define READKEY getch()
    #define PRINTF printw
    #define ERRPRINTF(arg) printw("CHYBA: "); printw(arg)
    #define SCANF scanw
    #define CLEAR clear()
    #define GOTO00 move(0,0);
#else
    #define READKEY getchar()
    #define PRINTF printf
    #define ERRPRINTF(arg) fprintf(stderr,arg)
    #define SCANF scanf
    #define CLEAR system(CLEARCMD)
	#ifdef OSLINUX
		#define GOTO00 printf("\033[0;0f");
	#else
		#define GOTO00 CLEAR
	#endif
#endif

/* --- Používané strutkurované datové typy ----------------------------------*/
typedef enum {  //výsledek provedení tahu (důleitý pro další nakládání s ním)
    moved=0,
    nomove=1,
    finishFound=10,
    quit=11,
} moveResult;

typedef enum {  //typ algoritmu generování
    fullRand=0,
    beginRand=1,
    fullStatic=2
} algType;

typedef enum {  //pojmenované směry
    left,
    top,
    right,
    bottom
} direction;

typedef union { //argument předávaný metodě pro generování labyrintu v závislosti na použitém algoritmu
    int permutations;
    direction* directionsOrder;
} algArg;

#ifdef USE_CURSES
    typedef enum {
        startColor=1,
        playerColor=2,
        finishColor=3,
        wallColor=4,
        emptyColor=5,
        visitedColor=6
    } pointColor;
#endif
/* --- Namespace gameParams, ------------------------------------------------*/
//Parametry hry, které lze změnit argumenty programu, nebo v nastavení.
namespace gameParams {
    bool onlyOneGame;
    int width;  //velikost labyrintu
    int height;

    //velikost terminálu
    int termWidth;
    int termHeight;

    bool visible;   //má se labyrint zobrazovat, nebo zůstat skryt a komunikovat s hráčem jen prostřednictvím textových zpráv?
    bool dropSeeds; //má při cestě zahazovat semínka?

    char empty; //znaky k vykreslení
    char visited;
    char wall;
    char start;
    char player;
    char finish;

    bool startOnLeft;   //má být start vlevo a cíl vpravo, nebo náhodně poházené po labyrintu?
    bool finishOnRight;

    algType algorithm;  //algoritmus generování

    //následující 2 mají význam prouze pro fullRand a beginRand algoritmy
    int seed;       //semínko náhodného generátoru
    int permutations;   //počet permutování

    //má význam pouze pro fullStatic algoritmus
    direction dimsPerms[DIRECTC];   //posloupnost směrů, jak se mají postupně zkoušet porcházet

    int shortcuts;  //počet propojů mezi cestami
}

#endif
