/*****************************************************************************/
/*
    Anny's Lab 1.0
    made by m@rtlin, VII-IX 2012
    martlin@seznam.cz, http://www.martlins-web.tode.cz/
    Licence: GNU/GPL, Open Source

    Multiplatformní (primárně určena pro GNU/Linux) textová bludišťovka volitelně využívající knihovnu ncurses
    Přeji Hodně zábavy!
*/
/*
    Soubor map.cpp:

    Implementace třídy map reprezentující samotný labyrint.
*/
/*****************************************************************************/
#ifndef _al_map
#define _al_map

#include <cstdlib>
#include <cstdio>
#include <ctime>
#include <stack>

#include "declarations.cpp"
#include "fundec.cpp"
#include "others.cpp"
#include "points.cpp"
#include "twodimarr.cpp"

/*****************************************************************************/
/** Trida Map. obsahuje mapu labiryntu, pozice startu, cile a hrace a dalsich par chujovin*/
class Map : public TwoDimArr {
    private:
        char wall;  //sloty znakové
        char empty;
        char visited;

        DrawablePoint *start;   //sloty pozicové
        DrawablePoint *finish;
        DrawablePoint *players;

        bool visible;   //sloty doplňujícho nastavení
        bool playable;

        time_t startTime;   //sloty ostatní
        int totalEggs;
        int foundEggs;


        int randBoardPos(int from, int (*fn)(int));   //x/y náhodná souřadnice v labyrintu

        void movePlayers(Point toPoint);
        void moveToPlayers(Point toPoint);

        void init(void);

        Point* createDirsOrder(Point from, algType algoritm, algArg genParams);
        bool labGenIter(algType algoritm, algArg genParams);

        typedef struct {    //struktura ukládaná na zásobník při generování labyrintu
            int i;
            Point from;
            Point *to;
        } iterEnv;

    public:
        //konstrkturoy
        Map(void);  //bere všechny argumenty z gameParams
        Map(int seed);  //velikost z gameParams, vlastní číslo labyrintu-seed
        Map(int width, int height);  //potlačí velikost bludiště z gamePaams a použije tuto

        char getWall(void); //zapouzdreni - cteni
        char getEmpty(void);
        char getVisited(void);

        DrawablePoint getStart(void);
        DrawablePoint getFinish(void);
        DrawablePoint getPlayers(void);

        bool isPlayable(void);
        bool isVisible(void);

        time_t getStartTime(void);
        //zapouzdreni pro totoalEggs a foundEggs nebudou implemtnovany - tezkej private

        bool onBorder(Point p);   //dalsi metody
        Point randPointAround(Point p);
        void print(void);   //prepsana metoda pro vypis

        bool play(void);
        void help(void);
        void info(void);

              //generovani labyrintu
        bool createLabirynth(bool onlyBgnRnd, int permutations);   // nějakým způsobem náhodný algoritmus
        //+ statický algoritmus generování...odebrán pro nepoužitelnost

            //pohyb v labyrintu
        moveResult readMove(void);
        moveResult doMove(Point toPoint);
        ~Map(void);
};


/*****************************************************************************/
/* --- Konstruktory -------------------------------------------------------- */
Map::Map(void) : TwoDimArr(odd(gameParams::width), odd(gameParams::height), gameParams::wall){
    init();//o nastavení velikosti se stará konstruktor TwoDimArr
}

Map::Map(int seed) : TwoDimArr(odd(gameParams::width), odd(gameParams::height), gameParams::wall){
    start=players=finish=NULL;
    init();//o nastavení velikosti se stará konstruktor TwoDimArr
}

Map::Map(int width, int height) : TwoDimArr(odd(width), odd(height), gameParams::wall){
    init(); //o nastavení velikosti se stará konstruktor TwoDimArr
}
/* --- Destruktor ----------------------------------------------------------- */
Map::~Map(void) {
    playable=false;
    if (start) delete start;
    if (players) delete players;
    if (finish) delete finish;
}

void Map::init(void) {
    Point starts, finishs;
    bool genSucc;
    int i;

    start=players=finish=NULL;  //inicializace slotů na nulovou hodnotu, pro příad nevydařeného vygenerování

    CLEAR;
    if (IS_MAP_SMALL) {
        ERRPRINTF("Takto maly labyrint nejde vytvorit. Labyrint nebyl vytvoren.\n");
        READKEY;
        playable=false;
        return;
    }

    PRINTF("Probiha generovani Labyrintu. Vyckejte prosim.\n");
    #ifdef USE_CURSES
        refresh();
    #endif

    wall=gameParams::wall;
    empty=gameParams::empty;
    if (gameParams::dropSeeds)
        visited=gameParams::visited;
    else
        visited=gameParams::empty;
    visible=gameParams::visible;

    foundEggs=0;
    totalEggs=0;

    srand(gameParams::seed);

    start=new DrawablePoint(gameParams::startOnLeft ? 0 : randBoardPos(getSize().getX(), even),
                            randBoardPos(getSize().getY(), even),
                            gameParams::start);
    players=new DrawablePoint(getStart(), gameParams::player);
    finish=new DrawablePoint(gameParams::finishOnRight ? getSize().getX()-1 : randBoardPos(getSize().getX(), even),
                                randBoardPos(getSize().getY(), even),
                                gameParams::finish);
    while (getStart()==getFinish())  //pokud se cíl vygeneroval na stejné místo jako start, tak ho opět náhodně umisťuj, dokud nebude jinde
        finish->moveTo(gameParams::finishOnRight ? getSize().getX()-1 : randBoardPos(getSize().getX(), even),
                                randBoardPos(getSize().getY(), even));


/*
    PRINTF("\nMapa: <%d, %d>, S:<%d, %d>, C:<%d %d>\n",getSize().getX(), getSize().getY(), getStart().getX(), getStart().getY(),  getFinish().getX(), getFinish().getY());
    refresh();
    READKEY;
*/
    setItem(getStart());
    setItem(getFinish());

    switch (gameParams::algorithm) { //samotné vygenerování algoritmu
        case fullRand:
            genSucc=createLabirynth(false, gameParams::permutations);  break;
        case beginRand:
            genSucc=createLabirynth(true, gameParams::permutations);  break;
        /*
        case fullStatic:
            genSucc=createLabirynth(gameParams::dimsPerms[0], gameParams::dimsPerms[1], gameParams::dimsPerms[2], gameParams::dimsPerms[3]);   break;
        */
        default:
            ERRPRINTF("Tento algoritmus neni podporovan. Pouzijte jiny, nebo modofikujte program.");
            playable=false;
            return;
    }

    if (!genSucc) {  //bylo genrování neúspěšné?
        ERRPRINTF("Z nezname priciny se nepodarilo vytvorit labyrint.\n");
        READKEY;
        playable=false;
        return;
    } else
        playable=true;

    for (i=0; i<gameParams::shortcuts; i++) {    //do vytvořeného labyritu napráskáme pár propojů
        Point randPoint;
        bool horizOrVert=(rand()%2)==1;    //má to být probourání horizontální nebo vertiální stěny?

        if (horizOrVert)    //horizontální průraz
            randPoint.moveTo(randBoardPos(getSize().getX(), even), randBoardPos(getSize().getY(), odd));
        else    //vertikální průraz
            randPoint.moveTo(randBoardPos(getSize().getX(), odd), randBoardPos(getSize().getY(), even));
        setItem(randPoint, getEmpty());

    }
    setItem(getPlayers());  //umísti do labyrintu hráče a může se hrát!
}

/* --- Zapouzdreni - cteni ------------------------------------------------- */
char Map::getWall(void){
    return wall;
}

char Map::getEmpty(void) {
    return empty;
}

char Map::getVisited(void) {
    return visited;
}

DrawablePoint Map::getStart(void) {
    return *start;
}

DrawablePoint Map::getFinish(void) {
    return *finish;
}

DrawablePoint Map::getPlayers(void) {
    return *players;
}

bool Map::isPlayable(void) {
    return playable;
}

bool Map::isVisible(void) {
    return visible;
}

time_t Map::getStartTime(void) {
    return startTime;
}

/* --- Zapouzdreni - nastaveni --------------------------------------------- */
void Map::movePlayers(Point toPoint) {
    players->move(toPoint);
}

void Map::moveToPlayers(Point toPoint) {
    players->moveTo(toPoint);
}
/* --- Pomocne metody ------------------------------------------------------- */
bool Map::onBorder(Point p) {   //vraci true, je-li bod na ramecku
    return p.getX()==0 || p.getY()==0 || p.getX()==(getSize().getX()-1) || p.getY()==(getSize().getY()-1);
}

Point Map::randPointAround(Point p) {   //vrati nahodny bod v okoli bodu p, tak, aby byl validne v mape
    Point res;
    int x,y;
    do {
        x=((rand()%4+1)/2-1)*2; //náhodné číslo z množiny {-2,0,0,+2}
        if (x==0)
            y=(rand()%2)*4-2; //náhodné číslo z množiny {-2, +2}
        else
            y=0;
        res.moveTo(p.getX()+x, p.getY()+y);
    } while (!inBounds(res) || onBorder(res));
    return res;
}

int Map::randBoardPos(int from, int (*fn)(int)) { //vypočítá z meze from hodnotu v labyrintu
    if (fn==even)
        return even(rand()%(from-4))+2;
    else if (fn==odd)
        return even(rand()%(from-6))+3;
    else
        return fn(rand()%(from-4))+2;
}

void Map::print(void) { //vytiskne labyrint se vsim vsudy
    int left, top, right, bottom;
    int i,j;

    #ifdef USE_CURSES
        gameParams::termWidth=COLS-1;
        gameParams::termHeight=LINES-1;
    #endif

    left=getPlayers().getX()-gameParams::termWidth/2;    //standartni vypocet mezí tisku
    top=getPlayers().getY()-gameParams::termHeight/2;
    right=getPlayers().getX()+gameParams::termWidth/2+gameParams::termWidth%2;
    bottom=getPlayers().getY()+gameParams::termHeight/2+gameParams::termHeight%2;

    if (left<0) {   //a ořežez mezí
        left=0;
        right=gameParams::termWidth;
    }
    if (top<0) {
        top=0;
        bottom=gameParams::termHeight;
    }
    if (right>=getSize().getX()) {
        right=getSize().getX()-1;
        left=right-gameParams::termWidth;
    }
    if (bottom>=getSize().getY()) {
        bottom=getSize().getY()-1;
        top=bottom-gameParams::termHeight;
    }

    if (left<0) left=0;
    if (top<0) top=0;
    if (right>getSize().getX()) right=getSize().getX();
    if (bottom>getSize().getY()) bottom=getSize().getY();

    //vypsání je potřeba provést transponovaně protože pole máme uloženo po sloupcích
    for (j=top; j<=bottom; j++) {
        for (i=left; i<=right; i++){
            #ifdef USE_CURSES
                Point here(i,j);
                if (here==getPlayers()) //hráč se musí testova jako první, tenm á nejvyšší prioritu
                    addch(getPlayers().getCharToDraw() | COLOR_PAIR(playerColor));
                else if (here==getStart())
                    addch(getStart().getCharToDraw() | COLOR_PAIR(startColor));
                else if (here==getFinish())
                    addch(getFinish().getCharToDraw() | COLOR_PAIR(finishColor));
                else if (getItem(here)==getWall())
                    addch(getWall() | COLOR_PAIR(wallColor));
                else if (getItem(here)==getEmpty())
                    addch(getEmpty() | COLOR_PAIR(emptyColor));
                else if (getItem(here)==getVisited())
                    addch(getVisited() | COLOR_PAIR(visitedColor));
                else
                    PRINTF("%c", getItems()[i][j]);
            #else
                PRINTF("%c", getItems()[i][j]);
            #endif
        }


        #ifdef USE_CURSES
            if (gameParams::termWidth>=getSize().getX())   //jinak přetéká a vykrelsuje zbytečné řádky, nebo se naopak posuová
                PRINTF("\n");
        #else
            PRINTF("\n");
        #endif
    }
    //PRINTF("L: %d, R: %d, T: %d, B: %d. (Term: %dx%d, Map: %dx%d)\n",
    //       left,right,top, bottom, gameParams::termWidth, gameParams::termHeight, getSize().getX(), getSize().getY());

    #ifdef USE_CURSES
        refresh();
    #endif
}

void Map::help(void) {  //rychla napoveda
    PRINTF("\rOvladani: AWSD/4826 - pohyb, X/Q - konec, H - napoveda, I - rychle info. ");
    if (!isVisible())
        PRINTF("N - zobraz labyrint. ");

    if (isVisible()) {
        myFflush(stdin);
        READKEY;
    }
}

void Map::info(void) {  //rychle info o suradnicich
    PRINTF("\rLabyrint: %dx%d   Start: (%d,%d)   Vy: (%d,%d)   Cil: (%d,%d) ",
           getSize().getX(), getSize().getY(),  getStart().getX(), getStart().getY(), getPlayers().getX(), getPlayers().getY(), getFinish().getX(), getFinish().getY());

    if (isVisible()) {
        myFflush(stdin);
        READKEY;
    }
}

bool Map::createLabirynth(bool onlyBgnRnd, int permutations) {  //vytvori nejakym zpusobem nahodny labyrint
    Point from, to;
    algArg genParam;
    bool genRes;

    from.moveTo(getStart());
    to.moveTo(randPointAround(from));   //první krok

    if (onlyBgnRnd) {   //pokud má být posloupnost náhodná jen na začátku, vytvoř nějakou náhodnou posloupnost směrů, a předávej ji jako zadanou
        int i;
        genParam.directionsOrder=new direction[DIRECTC];
        if (!genParam.directionsOrder)
            return false;

        for (i=0; i<DIRECTC; i++) {
            genParam.directionsOrder[i]=(direction)i;   //nejdříve zaplníme ve správném pořadí
        }

        for (i=0; i<gameParams::permutations; i++) {  //náhodně pospřehazujeme
            direction tmp;
            int a=rand()%DIRECTC, b=rand()%DIRECTC;
            tmp=genParam.directionsOrder[a];
            genParam.directionsOrder[a]=genParam.directionsOrder[b];
            genParam.directionsOrder[b]=tmp;
        }
    } else {    //jinak zádné pořadí směrů nebude, bude totálně náhodné, takže místo něj předej null
        genParam.permutations=permutations;
    }

    genRes=labGenIter((onlyBgnRnd ? beginRand : fullRand), genParam);  //samotné vygenerování labyrintu. Bylo úspěšné?

    if (onlyBgnRnd)
        delete[] genParam.directionsOrder;

    return genRes;
}

bool Map::labGenIter(algType algoritm, algArg genParams) {  //vrací true, pokud se našla cesta do cíle při samotném generování layrintu

    std::stack<iterEnv*> points;
    iterEnv *actEnv;
    Point toAct, between;
    bool finishFound=false;

    //inicializace výchozího prostředí - start
    actEnv=new iterEnv;
    actEnv->i=-1;    //hned na začátku každého cyklu je inkrementace, proto -1
    actEnv->from=getStart();
    actEnv->to=createDirsOrder(getStart(), algoritm, genParams);
    points.push(actEnv);

    while (!points.empty()) {
        actEnv=points.top();
        actEnv->i++;

        if ((actEnv->i)>=DIRECTC) { //prošel-li jsi celé prostředí, vrať se o jedno výš
            actEnv=points.top();
            points.pop();
            delete[] actEnv->to;
            delete actEnv;
            continue;
        }

        toAct=actEnv->to[actEnv->i];    //uložíme si bod, kam se nyní chceme vydat jako toAct
        between.moveTo((actEnv->from.getX()+toAct.getX())/2,(actEnv->from.getY()+toAct.getY())/2);

        if (toAct==getFinish() && !finishFound) { //jsi v cíli poprvé (nejdřív je potřeba testovat cíl, až pak hranice labyrintu (může být mimo ně)
            setItem(between, getEmpty());    //prokopej cestu mezi aktuálním a cílem
            finishFound=true;
            continue;
        }

        if (!inBounds(toAct) || onBorder(toAct) || getItem(toAct)!=getWall())
            //jsi-li mimo labyrint nebo na kraji labyrintu, chceš-li jít do již prokopaného pole
            continue;

        //prokopeme cestu do nového bodu, a pomalu se do něj přesuneme
        setItem(toAct, getEmpty());
        setItem(between, getEmpty());

        //jsme tedy v novém bodě a tak si vytvoříme nové prostředí
        actEnv=new iterEnv;
        actEnv->i=-1;
        actEnv->from=toAct;
        actEnv->to=createDirsOrder(toAct, algoritm, genParams);
        points.push(actEnv);
    }

    return finishFound; //našli jsme cíl?
}

Point* Map::createDirsOrder(Point from, algType algoritm, algArg genParams) {   //vytvoří z bodu pole bodů, se kterýmui sousedí a zpermutuje je
    Point* to=new Point[DIRECTC];
    int i;

    for (i=0; i<DIRECTC; i++)
        to[i]=from;

    if (algoritm==fullRand) {    //budeme vytvářet náhodné permutace ?...
            int a, b;   //uvažuji, že permutace bude alespoň 1
            Point tmp;
            to[0].move(-2,0);
            to[1].move(0,2);
            to[2].move(2,0);
            to[3].move(0,-2);
            for (i=0; i<genParams.permutations; i++) {  // náhodně pospřehazujeme ...
                a=rand()%DIRECTC;
                b=rand()%DIRECTC;

                tmp=to[a];
                to[a]=to[b];
                to[b]=tmp;
            }
        } else {  //nebo se použijí zadané?
            for (i=0; i<DIRECTC; i++)   //převod pole dimensionsů directionsOrder na pole Pointů to
                switch (genParams.directionsOrder[i]) {
                    case top:   to[i].move(0,-2);   break;
                    case right: to[i].move(2,0);    break;
                    case bottom:to[i].move(0,2);    break;
                    case left:  to[i].move(-2,0);   break;
                }
        }

    return to;
}

/* --- Hraní hry --------------------------------------------------------------*/
moveResult Map::doMove(Point toPoint) {

    if (getItem(toPoint)==getWall()) {   //do zdi tahnout nemuze
        if (!isVisible())
            PRINTF("\nZed!\n");
        else
            PRINTF("\nZed!\a\n");
        return nomove;
    }
    if (getItem(toPoint)=='0')
        foundEggs++;

    if (getPlayers()==getStart()) //nejdrive vypln puvodni pozici hrace startem nebo visited
        setItem(getPlayers(), getStart().getCharToDraw());
    else
        setItem(getPlayers(), getVisited());

    moveToPlayers(toPoint);    //samotny presun
    setItem(getPlayers());  //a vykresleni na nove pozici

    if (toPoint==getFinish()) {
        CLEAR;   //po dokončení se labyrint vytiskne automaticky, bez ohledu na to, jestli se zobrazi má nebo nemá
        print();
        myFflush(stdin);
        PRINTF("\rBlahopreji, jsi v cili!\n");
        READKEY;
        return finishFound;
    }
    return moved;
}

moveResult Map::readMove(void) {  //vraci true, pokud byla timto tahem hra ukonena
    char c;

    if (isVisible()) {
        CLEAR;
        print();
    }

    c=toUpperCase(READKEY);

    switch (c) {    //konec hry
        case 'X':
        case 'Q':   {
            return quit;
            break;
        }
        case 'H':   {   //nápověda
            help();
            return nomove;
            break;
        }
        case 'I': {     //info
            info();
            return nomove;
            break;
        }
        case 'C':   {   //cheat
            Point toPoint(rand()%getSize().getX(), rand()%getSize().getY());
            return doMove(toPoint);
        }
        case '0':
        case 'E':   {   //easter egg
            Point eggPos(rand()%getSize().getX(), rand()%getSize().getY());
            if (getItem(eggPos)==getEmpty() || getItem(eggPos)==getVisited())
                setItem(eggPos, '0');
                totalEggs++;
            return nomove;
        }

        case 'N': {     //zobrazení labyrintu v režimu skrytého labyrintu
            if (!isVisible()) {
                CLEAR;
                PRINTF("Atualni stav bludiste:\n");
                print();
            }
        } break;
        case 'A':   //doleva
        case '4': {
            Point toPoint=getPlayers();
            toPoint.move(-1,0);
            return doMove(toPoint);
            break;
        }
        case 'W':   //nahoru
        case '8': {
            Point toPoint=getPlayers();
            toPoint.move(0,-1);
            return doMove(toPoint);
            break;
        }
        case 'S':   //dolů
        case '2': {
            Point toPoint=getPlayers();
            toPoint.move(0,1);
            return doMove(toPoint);
            break;
        }
        case 'D':   //doprava
        case '6': {
            Point toPoint=getPlayers();
            toPoint.move(1,0);
            return doMove(toPoint);
            break;
        }
        default:
            return nomove;
    }
    return nomove;  //nemělo by nastat
}


bool Map::play(void) {  //vraci true, pokud byla hra ukončena nalezením cíle
    if (!isPlayable()) {
        ERRPRINTF("Hru nelze spustit. Labyrint neni hratelny.\n");
        return false;
    } else {
        moveResult result;
        startTime=time(NULL);   //spustí časomíru
        #ifdef USE_CURSES
            if (!isVisible())
                scrollok(stdscr, true);
            else
                scrollok(stdscr, false);
        #endif


        if (!isVisible())
            PRINTF("Nyni jste na startu (%d, %d). Dojdete do cile (%d, %d).\nVase kroky:\n", getPlayers().getX(), getPlayers().getY(), getFinish().getX(), getFinish().getY());

        do {
            result=readMove();
        } while (result!=quit && result!=finishFound);
        #ifdef USE_CURSES
            scrollok(stdscr, true);
        #endif

        if (result==finishFound) {
            CLEAR;
            PRINTF("Cas: %.0f sekund.\n", difftime(time(NULL),getStartTime()));
            if (foundEggs>0)
                PRINTF("Nalezeno %d z %d velikonocnich vajicek.\n", foundEggs, totalEggs);

            READKEY;
            return true;
        } else
            return false;
    }
}

#endif
